--
-- ReceivingHopper
--
-- @author  Manuel Leithner
-- @date  05/08/16
--
-- Copyright (C) GIANTS Software GmbH, Confidential, All Rights Reserved.
--
-- Edit for Fliegl PFS 16000 by Ifko[nator] @ 07.01.2017


source("dataS/scripts/vehicles/specializations/events/ReceivingHopperSetCreateBoxesEvent.lua");


FlieglPFS16000 = {};

function FlieglPFS16000.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Fillable, specializations);
end;

function FlieglPFS16000:load(savegame)
    self.setUnitFillLevel            = Utils.appendedFunction(self.setUnitFillLevel, FlieglPFS16000.setUnitFillLevel);
    self.createBox                   = FlieglPFS16000.createBox;
    self.onBoxTriggerCallback        = FlieglPFS16000.onBoxTriggerCallback;
    self.findFillableRaycastCallback = FlieglPFS16000.findFillableRaycastCallback;
    self.setCreateBoxes              = Utils.overwrittenFunction(self.setCreateBoxes, FlieglPFS16000.setCreateBoxes);
	self.getIsFoldAllowed 			 = Utils.overwrittenFunction(self.getIsFoldAllowed, FlieglPFS16000.getIsFoldAllowed);

    self.flieglPFS16000 = {};
    self.flieglPFS16000.unloadingDelay = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.flieglPFS16000#unloadingDelay"), 0) * 1000;
    self.flieglPFS16000.unloadingStartTime = -1;
    self.flieglPFS16000.fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.flieglPFS16000#fillUnitIndex"), 1);
    self.flieglPFS16000.unloadInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.flieglPFS16000#unloadInfoIndex"), 1);
    self.flieglPFS16000.dischargeInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.flieglPFS16000#dischargeInfoIndex"), 1);

    local triggerId = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.flieglPFS16000.tipTrigger#index"));
    
	if triggerId ~= nil and triggerId ~= 0 then
        self.flieglPFS16000.tipTrigger = ReceivingTipTrigger:new(self.isServer, self.isClient);
        self.flieglPFS16000.tipTrigger:load(triggerId, self);
        self.flieglPFS16000.tipTrigger:register(true)
        self.flieglPFS16000.tipTrigger:addUpdateEventListener(self)
    end;

    self.flieglPFS16000.boxSpawnPlace = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.flieglPFS16000.boxTrigger#boxSpawnPlaceIndex"));

    self.flieglPFS16000.boxTriggerNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.flieglPFS16000.boxTrigger#index"));
    
	if self.flieglPFS16000.boxTriggerNode ~= nil and self.isServer then
       addTrigger(self.flieglPFS16000.boxTriggerNode, "onBoxTriggerCallback", self);
    end;

    self.flieglPFS16000.isOverloading = false;
    self.flieglPFS16000.boxes = {};
	
    local boxesCount = 0;
    
	while true do
        local baseName = "vehicle.flieglPFS16000.boxTrigger.box(" .. tostring(boxesCount) .. ")";
        if not hasXMLProperty(self.xmlFile, baseName) then
            break;
        end;

        local fillTypeString = getXMLString(self.xmlFile, baseName.."#fillType");
        local filename = getXMLString(self.xmlFile, baseName.."#filename");
        local fillType = FillUtil.fillTypeNameToInt[fillTypeString];
        
		if fillType ~= nil then
            self.flieglPFS16000.boxes[fillType] = filename;
        end;
		
        boxesCount = boxesCount + 1;
    end;

    if self.isClient then
		self.flieglPFS16000.turnedOnRotationNodes = Utils.loadRotationNodes(self.xmlFile, {}, "vehicle.turnedOnRotationNodes.turnedOnRotationNode", "flieglPFS16000", self.components); 
        self.flieglPFS16000.fillScrollerNodes = Utils.loadScrollers(self.components, self.xmlFile, "vehicle.flieglPFS16000.fillScrollerNodes.fillScrollerNode", {}, false);
        self.flieglPFS16000.fillBoxEffects = EffectManager:loadEffect(self.xmlFile, "vehicle.flieglPFS16000.fillEffect", self.components, self);
        self.flieglPFS16000.currentBoxFillType = nil;
    end;

    self.flieglPFS16000.litersPerMs = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.flieglPFS16000.boxTrigger#litersPerMinute"), 2000) / 60 / 1000;
    self.flieglPFS16000.raycastNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.flieglPFS16000.raycastNode#index"));
    self.flieglPFS16000.raycastMaxDistance = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.flieglPFS16000.raycastNode#raycastLength"), 10);
    self.flieglPFS16000.objectsInBoxPlace = 0;
    self.flieglPFS16000.timeUntilNextBox = 3000;
    self.flieglPFS16000.nextBoxWaitTime = 3000;
    self.flieglPFS16000.currentBox = nil;
    self.flieglPFS16000.createBoxes = false;
	
	self.liquidTank = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.flieglPFS16000.liquidTank#index"));

	if self.liquidTank ~= nil then
		setVisibility(self.liquidTank, false);
	end;
	
	
	if self.sampleFill == nil then
        local linkNode = Utils.indexToObject(self.components, Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.fillSound#linkNode"), "0>"));
        self.sampleFill = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.fillSound", nil, self.baseDirectory, linkNode);
    end;
    
	self.sampleFillEnabled = false;
    self.sampleFillStopTime = -1;
    self.lastFillLevel = -1;
end;

function FlieglPFS16000:getIsFoldAllowed(superFunc, onAiTurnOn)
   return self.turnOnVehicle and self:getIsTurnedOn();
end

function FlieglPFS16000:delete()
    if self.flieglPFS16000.tipTrigger ~= nil then
        self.flieglPFS16000.tipTrigger:delete();
    end;

    if self.isServer then
        if self.flieglPFS16000.boxTriggerNode ~= nil then
            removeTrigger(self.flieglPFS16000.boxTriggerNode);
        end;
    end;

    if self.isClient then
        EffectManager:deleteEffects(self.flieglPFS16000.fillBoxEffects);
    end;
	
	if self.isClient then
        SoundUtil.deleteSample(self.sampleFill);
    end;
end;

function FlieglPFS16000:readStream(streamId, connection)
    if connection:getIsServer() then
		local tipTriggerId = streamReadInt32(streamId);
	
        self.flieglPFS16000.isOverloading = streamReadBool(streamId);
        self.flieglPFS16000.tipTrigger:readStream(streamId, connection);
        
		g_client:finishRegisterObject(self.flieglPFS16000.tipTrigger, tipTriggerId);
    end;
end;

function FlieglPFS16000:writeStream(streamId, connection)
    if not connection:getIsServer() then
        streamWriteBool(streamId, self.flieglPFS16000.isOverloading);
        streamWriteInt32(streamId, self.flieglPFS16000.tipTrigger.id);
        
		self.flieglPFS16000.tipTrigger:writeStream(streamId, connection);
        
		g_server:registerObjectInStream(connection, self.flieglPFS16000.tipTrigger);
    end;
end;

function FlieglPFS16000:readUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        self.flieglPFS16000.isOverloading = streamReadBool(streamId);
    end;
end;

function FlieglPFS16000:writeUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        streamWriteBool(streamId, self.flieglPFS16000.isOverloading)
    end;
end;

function FlieglPFS16000:mouseEvent(posX, posY, isDown, isUp, button)
end;

function FlieglPFS16000:keyEvent(unicode, sym, modifier, isDown)
end;

function FlieglPFS16000:update(dt)
    if self.isClient then
		Utils.updateRotationNodes(self, self.flieglPFS16000.turnedOnRotationNodes, dt, self.flieglPFS16000.isOverloading);
	
		 if self.turnOnVehicle and self:getIsTurnedOn() then
			if self:getIsActiveForInput() then
				if InputBinding.hasEvent(InputBinding.TOGGLE_TIPSTATE) then
					self:setCreateBoxes(not self.createBoxes);
				end;
			end;
		
			if self.fillUnits ~= nil and self.flieglPFS16000.isOverloading then
				for _, fillUnit in pairs(self.fillUnits) do
                    local fillVolume = self.fillVolumes[fillUnit.fillVolumeIndex];
                    
					if fillVolume ~= nil then
                        if fillVolume.scrollSpeedLoad[1] ~= 0 or fillVolume.scrollSpeedLoad[2] ~= 0 or fillVolume.scrollSpeedLoad[3] ~= 0 then
                            fillVolume.uvPosition[1] = fillVolume.uvPosition[1] + fillVolume.scrollSpeedLoad[1] * dt;
                            fillVolume.uvPosition[2] = fillVolume.uvPosition[2] + fillVolume.scrollSpeedLoad[2] * dt;
                            fillVolume.uvPosition[3] = fillVolume.uvPosition[3] + fillVolume.scrollSpeedLoad[3] * dt;
                            
							setShaderParameter(fillVolume.volume, "uvOffset", fillVolume.uvPosition[1], fillVolume.uvPosition[2], fillVolume.uvPosition[3], 0, false);
                        end;
                    end;
                end;
            end;
		end;
    end;
	
	if (self.fillUnits and self.liquidTank) ~= nil then
		if self:getFillLevel() > 0 then	
			for _, fillUnit in pairs(self.fillUnits) do
				setVisibility(self.liquidTank, FillUtil.categoryToFillTypes[FillUtil.FILLTYPE_CATEGORY_LIQUID][fillUnit.currentFillType] ~= nil);
			end;
		else
			setVisibility(self.liquidTank, false);
		end;
	end;
end;

function FlieglPFS16000:updateTick(dt)
    if self.isServer then
        self.flieglPFS16000.isOverloading = false;
        
		local fillLevel = self:getUnitFillLevel(self.flieglPFS16000.fillUnitIndex);
        
		if fillLevel > 0 then
			if self.turnOnVehicle and self:getIsTurnedOn() then
				local fillType = self:getUnitFillType(self.flieglPFS16000.fillUnitIndex);
				
				if self.flieglPFS16000.currentBox ~= nil then
					if self.flieglPFS16000.currentBox:getFillLevel() < self.flieglPFS16000.currentBox:getCapacity() then
						local delta = math.min(math.min(self.flieglPFS16000.litersPerMs * dt, fillLevel), self.flieglPFS16000.currentBox:getCapacity() - self.flieglPFS16000.currentBox:getFillLevel());
						
						self.flieglPFS16000.currentBox:setFillLevel(self.flieglPFS16000.currentBox:getFillLevel() + delta);
						self:setUnitFillLevel(self.flieglPFS16000.fillUnitIndex, fillLevel - delta, fillType, false, self.fillVolumeUnloadInfos[self.flieglPFS16000.unloadInfoIndex]);
						self.flieglPFS16000.isOverloading = true;
					end;
				else
					if self.flieglPFS16000.objectsInBoxPlace == 0 then
						self.fillableObject = nil;
						
						if self.flieglPFS16000.raycastNode ~= nil then
							local x,y,z = getWorldTranslation(self.flieglPFS16000.raycastNode);
							local dx,dy,dz = localDirectionToWorld(self.flieglPFS16000.raycastNode, 0,-1,0);
							
							raycastAll(x,y,z, dx,dy,dz, "findFillableRaycastCallback", self.flieglPFS16000.raycastMaxDistance, self);
						end
	
						if self.fillableObject == nil then
							self.flieglPFS16000.timeUntilNextBox = self.flieglPFS16000.timeUntilNextBox - dt;
							
							if self.flieglPFS16000.timeUntilNextBox < 0 then
								if g_currentMission:getCanAddLimitedObject(FSBaseMission.LIMITED_OBJECT_TYPE_PALLET) then
									self:createBox();
								else
									if self.createBoxes then
										g_currentMission:showBlinkingWarning(g_i18n:getText("warning_tooManyPallets"), 2000);
									end;
								end;
								
								self.flieglPFS16000.timeUntilNextBox = self.flieglPFS16000.nextBoxWaitTime;
							end;
						else
							local object = self.fillableObject;
							
							if object:allowFillType(fillType) then
								if self.flieglPFS16000.unloadingStartTime < 0 then
									self.flieglPFS16000.unloadingStartTime = g_currentMission.time;
								end;
	
								local freeCapacity = object:getCapacity(fillType) - object:getFillLevel(fillType);
								
								if freeCapacity > 0 then
									self.flieglPFS16000.isOverloading = true;
								end;
	
								if self.flieglPFS16000.unloadingStartTime + self.flieglPFS16000.unloadingDelay < g_currentMission.time then
									local delta = math.min(fillLevel, freeCapacity, self.flieglPFS16000.litersPerMs * dt);
									
									if math.abs(delta) > 0 then
										self:setUnitFillLevel(self.flieglPFS16000.fillUnitIndex, fillLevel - delta, fillType, false, self.fillVolumeUnloadInfos[self.flieglPFS16000.unloadInfoIndex]);
										object:setFillLevel(object:getFillLevel(fillType) + delta, fillType, false, self.fillVolumeDischargeInfos[self.flieglPFS16000.dischargeInfoIndex]);
									else
										self.flieglPFS16000.unloadingStartTime = -1;
									end
								end;
							end
						end
					else
						self.flieglPFS16000.timeUntilNextBox = self.flieglPFS16000.nextBoxWaitTime;
					end;
				end;
			end;
        end;
    end;

    if self.isClient then
		Utils.updateScrollers(self.flieglPFS16000.fillScrollerNodes, dt, self.flieglPFS16000.isOverloading, self.flieglPFS16000.isOverloading);

        if self.flieglPFS16000.fillBoxEffects ~= nil then
            if self.flieglPFS16000.isOverloading then
                --local lastValidFillType = self:getUnitLastValidFillType(self.flieglPFS16000.fillUnitIndex);
                --EffectManager:setFillType(self.flieglPFS16000.fillBoxEffects, lastValidFillType)
                --EffectManager:startEffects(self.flieglPFS16000.fillBoxEffects)
            else
                EffectManager:stopEffects(self.flieglPFS16000.fillBoxEffects)
            end;
        end;
    end;
	
	
	if self.isClient then
        if self.isFilling then
            if self:getIsActiveForSound(true) then
                SoundUtil.playSample(self.sampleFill, 0, 0, nil);
                SoundUtil.stop3DSample(self.sampleFill);
            else
                SoundUtil.stopSample(self.sampleFill);
                SoundUtil.play3DSample(self.sampleFill);
            end;
        else
            SoundUtil.stopSample(self.sampleFill);
            SoundUtil.stop3DSample(self.sampleFill);
        end;
    end;
end;

function FlieglPFS16000:draw()
    if self.isClient then
        if self:getIsActiveForInput() and self.turnOnVehicle and self:getIsTurnedOn() then
            if self.createBoxes then
                g_currentMission:addHelpButtonText(g_i18n:getText("action_disablePalletSpawning"), InputBinding.TOGGLE_TIPSTATE, nil, GS_PRIO_NORMAL);
            else
                g_currentMission:addHelpButtonText(g_i18n:getText("action_enablePalletSpawning"), InputBinding.TOGGLE_TIPSTATE, nil, GS_PRIO_NORMAL);
            end;
        end;
    end;
end;

function FlieglPFS16000:setUnitFillLevel(fillUnitIndex, fillLevel, fillType, force, fillInfo)
    if fillLevel > 0 then
        if self.isClient then
            if self.flieglPFS16000.fillBoxEffects ~= nil then
                if fillType ~= self.flieglPFS16000.currentBoxFillType then
                    EffectManager:stopEffects(self.flieglPFS16000.fillBoxEffects);
                end;
            end;
            
			self.flieglPFS16000.currentBoxFillType = fillType;
        end;
    end;
end;

function FlieglPFS16000:setCreateBoxes(superFunc, state, noEventSend)
    if superFunc ~= nil then
        superFunc(self, state, noEventSend);
    end;

    ReceivingHopperSetCreateBoxesEvent.sendEvent(self, state, noEventSend);
    self.createBoxes = state;
end;

function FlieglPFS16000:createBox()
    if self.isServer then
		if self.turnOnVehicle and self:getIsTurnedOn() then
			if self.createBoxes then
				local fillType = self:getUnitFillType(self.flieglPFS16000.fillUnitIndex);
				local i3dFilename = Utils.getFilename(self.flieglPFS16000.boxes[fillType], self.baseDirectory);
				local x,y,z = getWorldTranslation(self.flieglPFS16000.boxSpawnPlace);
				local rx,ry,rz = getWorldRotation(self.flieglPFS16000.boxSpawnPlace);
				
				self.flieglPFS16000.currentBox = FillablePallet:new(self.isServer, self.isClient);
				self.flieglPFS16000.currentBox:load(i3dFilename, x,y,z, rx,ry,rz);
				self.flieglPFS16000.currentBox:register();
			end;
        end;
    end
end;

function FlieglPFS16000:onBoxTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter then
        local object = g_currentMission:getNodeObject(otherShapeId);
        
		if object ~= nil then
            self.flieglPFS16000.objectsInBoxPlace = self.flieglPFS16000.objectsInBoxPlace + 1;
            
			local fillType = self:getUnitFillType(self.flieglPFS16000.fillUnitIndex);
            
			if self.flieglPFS16000.currentBox == nil and object:isa(FillablePallet) and object:getFillType() == fillType and object:getFillLevel() < object:getCapacity() then
                self.flieglPFS16000.currentBox = object;
            end;
        end;
    elseif onLeave then
        local object = g_currentMission:getNodeObject(otherShapeId);
        
		if object ~= nil then
            self.flieglPFS16000.objectsInBoxPlace = math.max(0, self.flieglPFS16000.objectsInBoxPlace - 1);
        end;
        
		if object == self.flieglPFS16000.currentBox then
            self.flieglPFS16000.currentBox = nil;
        end;
    end;
end;

function FlieglPFS16000:findFillableRaycastCallback(transformId, x, y, z, distance)
    local vehicle = g_currentMission.nodeToVehicle[transformId];
    if vehicle ~= nil then
        if vehicle.exactFillRootNode == transformId then
            self.fillableObject = vehicle;
            
			return false;
        elseif vehicle.shovel ~= nil then
            self.fillableObject = vehicle;
            
			return false;
        end;
    end;
    
	return true;
end;